@@ -1,6 +1,6 @@
 /* -*- c-file-style: "ruby"; indent-tabs-mode: nil -*- */
 /*
- *  Copyright (C) 2011  Ruby-GNOME2 Project Team
+ *  Copyright (C) 2011-2015  Ruby-GNOME2 Project Team
  *  Copyright (C) 2002-2004  Ruby-GNOME2 Project Team
  *  Copyright (C) 2002-2003  Masahiro Sakai
  *  Copyright (C) 1998-2000 Yukihiro Matsumoto,
@@ -170,22 +170,12 @@ rbgobj_init_object_class(VALUE klass)
 
 /**********************************************************************/
 
-static gboolean
-is_gtkobject(GObject *gobj)
-{
-    static GType gtype_gtkobject = G_TYPE_INVALID;
-    if (!gtype_gtkobject)
-        gtype_gtkobject = g_type_from_name("GtkObject");
-    return gtype_gtkobject && g_type_is_a(G_OBJECT_TYPE(gobj), gtype_gtkobject);
-}
-
 static void
 gobj_mark(gpointer ptr)
 {
     GObject* gobj = ptr;
     guint n_properties;
     GParamSpec** properties;
-    GValue gval = G_VALUE_INIT;
     guint i;
 
     properties = g_object_class_list_properties(G_OBJECT_GET_CLASS(gobj), &n_properties);
@@ -198,6 +188,7 @@ gobj_mark(gpointer ptr)
         /* FIXME: exclude types that doesn't have identity. */
 
         {
+            GValue gval = G_VALUE_INIT;
             g_value_init(&gval, value_type);
             g_object_get_property(gobj, pspec->name, &gval);
             rbgobj_gc_mark_gvalue(&gval);
@@ -227,15 +218,7 @@ rg_s_new_bang(int argc, VALUE *argv, VALUE self)
 
     gobj = rbgobj_gobject_new(cinfo->gtype, params_hash);
     result = GOBJ2RVAL(gobj);
-
-    // XXX: Ughhhhh
-    if (is_gtkobject(gobj)){
-        // We can't call gtk_object_sink() here.
-        // But hopefully someone will call it in the future.
-        //gtk_object_sink(gobj);
-    } else {
-        g_object_unref(gobj);
-    }
+    g_object_unref(gobj);
 
     return result;
 }
@@ -319,7 +302,7 @@ rbgobj_gobject_new(GType gtype, VALUE params_hash)
         size_t param_size;
         struct param_setup_arg arg;
 
-        param_size = NUM2INT(rb_funcall(params_hash, rb_intern("length"), 0)); 
+        param_size = NUM2INT(rb_funcall(params_hash, rb_intern("length"), 0));
 
         arg.param_size = param_size;
         arg.gclass = G_OBJECT_CLASS(g_type_class_ref(gtype));
@@ -637,6 +620,137 @@ rg_type_name(VALUE self)
     return CSTR2RVAL(G_OBJECT_TYPE_NAME(RVAL2GOBJ(self)));
 }
 
+#if GLIB_CHECK_VERSION(2, 26, 0)
+typedef struct {
+    VALUE transform_from_callback;
+    VALUE transform_to_callback;
+    VALUE self;
+} RGBindPropertyCallbackData;
+
+static gboolean
+rg_bind_property_transform_to_callback(G_GNUC_UNUSED GBinding *binding,
+                                       const GValue *from_value,
+                                       GValue *to_value,
+                                       gpointer user_data)
+{
+    RGBindPropertyCallbackData *data = (RGBindPropertyCallbackData *)user_data;
+    VALUE proc;
+    VALUE rb_from_value;
+    VALUE rb_to_value;
+
+    proc = data->transform_to_callback;
+    if (NIL_P(proc))
+        return FALSE;
+
+    rb_from_value = rbgobj_gvalue_to_rvalue(from_value);
+    rb_to_value = rbgobj_gvalue_to_rvalue(to_value);
+    rb_to_value = rb_funcall(proc, rb_intern("call"), 1, rb_from_value);
+    rbgobj_rvalue_to_gvalue(rb_to_value, to_value);
+    return TRUE;
+}
+
+static gboolean
+rg_bind_property_transform_from_callback(G_GNUC_UNUSED GBinding *binding,
+                                         const GValue *from_value,
+                                         GValue *to_value,
+                                         gpointer user_data)
+{
+    RGBindPropertyCallbackData *data = (RGBindPropertyCallbackData *)user_data;
+    VALUE proc;
+    VALUE rb_from_value;
+    VALUE rb_to_value;
+
+    proc = data->transform_from_callback;
+    if (NIL_P(proc))
+        return FALSE;
+
+    rb_from_value = rbgobj_gvalue_to_rvalue(from_value);
+    rb_to_value = rbgobj_gvalue_to_rvalue(to_value);
+    rb_to_value = rb_funcall(proc, rb_intern("call"), 1, rb_from_value);
+    rbgobj_rvalue_to_gvalue(rb_to_value, to_value);
+    return TRUE;
+}
+
+static void
+rg_destroy_bind_property_full_data(gpointer user_data)
+{
+    RGBindPropertyCallbackData *data = (RGBindPropertyCallbackData *)user_data;
+
+    if (!NIL_P(data->transform_to_callback))
+        G_CHILD_REMOVE(data->self, data->transform_to_callback);
+
+    if (!NIL_P(data->transform_from_callback))
+        G_CHILD_REMOVE(data->self, data->transform_from_callback);
+
+    xfree(data);
+}
+
+static VALUE
+rg_bind_property(gint argc, VALUE *argv, VALUE self)
+{
+    VALUE rb_source_property;
+    VALUE rb_target;
+    VALUE rb_target_property;
+    VALUE rb_flags;
+    VALUE rb_options;
+    VALUE rb_transform_to;
+    VALUE rb_transform_from;
+
+    gpointer source;
+    const gchar *source_property;
+    gpointer target;
+    const gchar *target_property;
+    GBindingFlags flags;
+    GBinding *binding;
+    GBindingTransformFunc transform_to = NULL;
+    GBindingTransformFunc transform_from = NULL;
+
+    rb_scan_args(argc, argv, "41", &rb_source_property, &rb_target,
+                 &rb_target_property, &rb_flags, &rb_options);
+
+    rbg_scan_options(rb_options,
+                     "transform_to", &rb_transform_to,
+                     "transform_from", &rb_transform_from,
+                     NULL);
+
+    source = RVAL2GOBJ(self);
+    source_property = RVAL2CSTR(rb_source_property);
+    target = RVAL2GOBJ(rb_target);
+    target_property = RVAL2CSTR(rb_target_property);
+    flags = RVAL2GBINDINGFLAGS(rb_flags);
+
+    if (!NIL_P(rb_transform_to)) {
+        G_CHILD_ADD(self, rb_transform_to);
+        transform_to = rg_bind_property_transform_to_callback;
+    }
+
+    if (!NIL_P(rb_transform_from)) {
+        G_CHILD_ADD(self, rb_transform_from);
+        transform_from = rg_bind_property_transform_from_callback;
+    }
+
+    if (transform_to || transform_from) {
+        RGBindPropertyCallbackData *data;
+        data = (RGBindPropertyCallbackData *)xmalloc(sizeof(RGBindPropertyCallbackData));
+        data->self = self;
+        data->transform_to_callback = rb_transform_to;
+        data->transform_from_callback = rb_transform_from;
+        binding = g_object_bind_property_full(source, source_property,
+                                              target, target_property,
+                                              flags, transform_to,
+                                              transform_from,
+                                              (gpointer)data,
+                                              rg_destroy_bind_property_full_data);
+    } else {
+        binding = g_object_bind_property(source, source_property,
+                                         target, target_property,
+                                         flags);
+    }
+
+    return GOBJ2RVAL(binding);
+}
+#endif
+
 static VALUE
 rg_initialize(int argc, VALUE *argv, VALUE self)
 {
@@ -658,13 +772,6 @@ rg_initialize(int argc, VALUE *argv, VALUE self)
 
     gobj = rbgobj_gobject_new(RVAL2GTYPE(self), params_hash);
 
-    if (is_gtkobject(gobj)){
-        gobj = g_object_ref(gobj);
-        // We can't call gtk_object_sink() here.
-        // But hopefully someone will call it in the future.
-        //gtk_object_sink(gobj);
-    }
-
     G_INITIALIZE(self, gobj);
     return Qnil;
 }
@@ -679,7 +786,6 @@ gobj_ref_count(VALUE self)
 
 /**********************************************************************/
 
-static VALUE proc_mod_eval;
 static GQuark q_ruby_setter;
 static GQuark q_ruby_getter;
 
@@ -732,55 +838,51 @@ set_prop_func(GObject* object,
     rb_funcall(GOBJ2RVAL(object), ruby_setter, 1, GVAL2RVAL(value));
 }
 
-// FIXME: use rb_protect
-static void
-class_init_func(gpointer g_class_, G_GNUC_UNUSED gpointer class_data)
+void
+rbgobj_class_init_func(gpointer g_class, G_GNUC_UNUSED gpointer class_data)
 {
-    GObjectClass* g_class = G_OBJECT_CLASS(g_class_);
+    GObjectClass* g_object_class = G_OBJECT_CLASS(g_class);
 
-    g_class->set_property = set_prop_func;
-    g_class->get_property = get_prop_func;
-
-#if 0
-    VALUE class_init_proc = (VALUE)class_data;
-    rb_funcall(proc_mod_eval, rb_intern("call"), 2,
-               GTYPE2CLASS(G_TYPE_FROM_CLASS(g_class)), class_init_proc);
-#endif
+    g_object_class->set_property = set_prop_func;
+    g_object_class->get_property = get_prop_func;
 }
 
-static VALUE
-rg_s_type_register(int argc, VALUE* argv, VALUE self)
+void
+rbgobj_register_type(VALUE klass, VALUE type_name, GClassInitFunc class_init)
 {
-    VALUE type_name, flags;
-    volatile VALUE class_init_proc = Qnil;
     GType parent_type;
-    GTypeInfo* info;
-
-    rb_scan_args(argc, argv, "03", &type_name, &info, &flags);
+    GTypeInfo *info;
 
     {
-        const RGObjClassInfo* cinfo = rbgobj_lookup_class(self);
-        if (cinfo->klass == self)
-            rb_raise(rb_eTypeError, "already registered");
+        const RGObjClassInfo *cinfo = rbgobj_lookup_class(klass);
+        if (cinfo->klass == klass)
+            rb_raise(rb_eTypeError,
+                     "already registered class: <%s>",
+                     RBG_INSPECT(klass));
     }
 
     {
-        VALUE superclass = rb_funcall(self, rb_intern("superclass"), 0);
-        const RGObjClassInfo* cinfo = rbgobj_lookup_class(superclass);
+        VALUE superclass = rb_funcall(klass, rb_intern("superclass"), 0);
+        const RGObjClassInfo *cinfo = rbgobj_lookup_class(superclass);
         if (cinfo->klass != superclass)
-            rb_raise(rb_eTypeError, "super class must be registered to GLib");
+            rb_raise(rb_eTypeError,
+                     "super class must be registered to GLib: <%s>",
+                     RBG_INSPECT(superclass));
         parent_type = cinfo->gtype;
     }
 
     if (NIL_P(type_name)) {
-        VALUE s = rb_funcall(self, rb_intern("name"), 0);
+        VALUE klass_name = rb_funcall(klass, rb_intern("name"), 0);
 
-        if (strlen(StringValuePtr(s)) == 0)
-            rb_raise(rb_eTypeError, "can't determine type name");
+        if (strlen(StringValueCStr(klass_name)) == 0)
+            rb_raise(rb_eTypeError,
+                     "can't determine type name: <%s>",
+                     RBG_INSPECT(klass));
 
-        type_name = rb_funcall(
-            rb_eval_string("lambda{|x| x.gsub(/::/,'') }"),
-            rb_intern("call"), 1, s);
+        type_name = rb_funcall(klass_name, rb_intern("gsub"),
+                               2,
+                               rb_str_new_cstr("::"),
+                               rb_str_new_cstr(""));
     }
 
     {
@@ -793,9 +895,9 @@ rg_s_type_register(int argc, VALUE* argv, VALUE self)
         info->class_size     = query.class_size;
         info->base_init      = NULL;
         info->base_finalize  = NULL;
-        info->class_init     = class_init_func;
+        info->class_init     = class_init;
         info->class_finalize = NULL;
-        info->class_data     = (gpointer)class_init_proc;
+        info->class_data     = NULL;
         info->instance_size  = query.instance_size;
         info->n_preallocs    = 0;
         info->instance_init  = NULL;
@@ -804,37 +906,49 @@ rg_s_type_register(int argc, VALUE* argv, VALUE self)
 
     {
         GType type = g_type_register_static(parent_type,
-                                            StringValuePtr(type_name),
+                                            StringValueCStr(type_name),
                                             info,
-                                            NIL_P(flags) ? 0 : NUM2INT(flags));
-        G_RELATIVE(self, class_init_proc);
+                                            0);
 
-        rbgobj_register_class(self, type, TRUE, TRUE);
+        rbgobj_register_class(klass, type, TRUE, TRUE);
 
         {
-            RGObjClassInfo* cinfo = (RGObjClassInfo*)rbgobj_lookup_class(self);
+            RGObjClassInfo *cinfo = (RGObjClassInfo *)rbgobj_lookup_class(klass);
             cinfo->flags |= RBGOBJ_DEFINED_BY_RUBY;
         }
 
         {
             GType parent = g_type_parent(type);
-            const RGObjClassInfo* cinfo =  GTYPE2CINFO(parent);
-            VALUE m = rb_define_module_under(self, RubyGObjectHookModule);
+            const RGObjClassInfo *cinfo =  GTYPE2CINFO(parent);
+            VALUE initialize_module;
 
-            if (! (cinfo->flags & RBGOBJ_DEFINED_BY_RUBY)) {
-                rbg_define_method(m, "initialize", rg_initialize, -1);
+            initialize_module = rb_define_module_under(klass,
+                                                       RubyGObjectHookModule);
+            if (!(cinfo->flags & RBGOBJ_DEFINED_BY_RUBY)) {
+                rbg_define_method(initialize_module,
+                                  "initialize", rg_initialize, -1);
             }
 
-            rb_include_module(self, m);
+            rb_include_module(klass, initialize_module);
         }
-
-        return Qnil;
     }
 }
 
+static VALUE
+rg_s_type_register(int argc, VALUE *argv, VALUE self)
+{
+    VALUE type_name;
+
+    rb_scan_args(argc, argv, "01", &type_name);
+
+    rbgobj_register_type(self, type_name, rbgobj_class_init_func);
+
+    return Qnil;
+}
+
 /**********************************************************************/
 
-void 
+void
 Init_gobject_gobject(void)
 {
     RG_TARGET_NAMESPACE = G_DEF_CLASS_WITH_GC_FUNC(G_TYPE_OBJECT, "Object", mGLib,
@@ -869,6 +983,11 @@ Init_gobject_gobject(void)
     RG_DEF_METHOD(inspect, 0);
     RG_DEF_METHOD(type_name, 0);
 
+#if GLIB_CHECK_VERSION(2, 26, 0)
+    RG_DEF_METHOD(bind_property, -1);
+    G_DEF_CLASS(G_TYPE_BINDING_FLAGS, "BindingFlags", mGLib);
+#endif
+
     eNoPropertyError = rb_define_class_under(mGLib, "NoPropertyError",
                                              rb_eNameError);
 
@@ -879,7 +998,4 @@ Init_gobject_gobject(void)
 
     /* subclass */
     RG_DEF_SMETHOD(type_register, -1);
-
-    rb_global_variable(&proc_mod_eval);
-    proc_mod_eval = rb_eval_string("lambda{|obj,proc| obj.module_eval(&proc)}");
 }
